/*
Copyright (C) 1995 M. A. S. Elmohamed,
Northeast Parallel Architectures Center, Syracuse University.
saleh@npac.syr.edu
*/

/* ... testing stuff removed. 10/22/99  -Saleh */

#include <stdio.h>
#ifndef MSDOS
#include <strings.h>
#endif
#include <ctype.h>
#include "misc.h"
#include "config.h"
#include "pi.h"

/*---------------
** The epoch_base data holds seconds since midnight on sunday
** offsets to the days of the week.
**---------------*/
static ep_t epoch_base[] = {
  (0 * SECS_PER_DAY),		/* Sunday midnight (early sunday morning) */
  (1 * SECS_PER_DAY),		/* Monday midnight */
  (2 * SECS_PER_DAY),		/* Tuesday midnight */
  (3 * SECS_PER_DAY),		/* Wednesday midnight */
  (4 * SECS_PER_DAY),		/* Thursday midnight */
  (5 * SECS_PER_DAY),		/* Friday midnight */
  (6 * SECS_PER_DAY),		/* Saturday midnight */
};  

static char *week_day[7] = { "Sun","Mon","Tue","Wed","Thu","Fri","Sat" };


ft_t ft_create(ep_t start, ep_t stop, int assigned)
{
  ft_t ft = (ft_t) xcalloc(1,sizeof(ft_s));
  ft->bt = start;
  ft->et = stop;
  ft->assigned = assigned;

  return(ft);
}

ft_t ft_set(ft_t ft, ep_t start, ep_t stop, int assigned)
{
  ft->bt = start;
  ft->et = stop;
  ft->assigned = assigned;

  return(ft);
}

void ft_destroy(ft_t ft)
{
  xfree((char *)ft);
}

fta_t fta_create(void)
{
  fta_t fta = (fta_t)xcalloc(1,sizeof(fta_s));
  return(fta);
}

void fta_sort(fta_t fta)
{
  int done, i, ftalen = fta->nft;
  ft_t *ft = fta->ft, tft;
  
  done = 0;
  while (done == 0) {
    done = 1;
    for (i = 0; i < ftalen - 1; i++) {
      if (ft[i]->bt > ft[i+1]->bt) {
	tft = ft[i];
	ft[i] = ft[i+1];
	ft[i+1] = tft;
	done = 0;
      }
    }
  }
}

char *ft_time_string(int twelve_hour, ep_t bt, ep_t et, char *buf)
{
  conv_interval24_string(twelve_hour, bt, et, buf);
  return(buf);
}

ft_t fta_add_time(fta_t fta, ep_t start, ep_t stop, int assigned)
{
  ft_t ft = (ft_t)0;
  register int i;

  if (fta != (fta_t)0) {
    if (fta->nft >= fta->nft_limit) { /* need more array space */
      if (fta->ft == (ft_t *)0)
	fta->ft = (ft_t *)
	  xcalloc((fta->nft_limit + FTA_INCREMENT),sizeof(ft_t));
      else {
	fta->ft = (ft_t *)
	  xrealloc((char *)fta->ft,
		   (fta->nft_limit + FTA_INCREMENT) * sizeof(ft_t));
	  for (i = fta->nft_limit; i < fta->nft_limit + FTA_INCREMENT; i++)
	    fta->ft[i] = (ft_t)0;
	}
      fta->nft_limit += FTA_INCREMENT;
    }
    /* if space is not already allocated, then create a new elt */
    if (fta->ft[fta->nft] == (ft_t)0)
      ft = fta->ft[fta->nft] = ft_create(start, stop, assigned);
    else
      ft = ft_set(fta->ft[fta->nft],start,stop, assigned);
    fta->nft++;
  }
    
  return(ft);
}

/* optimization */
void fta_removeft(fta_t fta, int ix)
{
  int nft = fta->nft;
  ft_t ft;

  if (ix < nft - 1) {
    ft = fta->ft[ix];
    fta->ft[ix] = fta->ft[nft - 1];
    fta->ft[nft - 1] = ft;	/* can't just throw it away --> memory leak */
  }
  if (nft > 0)
    fta->nft--;
}

void fta_deleteft(fta_t fta, int ix)
{
  int nft = fta->nft;
  register int i;

  if (ix < nft) {
    ft_destroy(fta->ft[ix]);
    for (i = ix; i < nft - 1; i++)
      fta->ft[i] = fta->ft[i+1];
    fta->nft--;
  }
}

void fta_intersect(fta_t src_fta, fta_t dst_fta)
{
  int src_nft = fta_getlength(src_fta);
  ft_t* src_ft = fta_getft_array(src_fta);
  int dst_nft = fta_getlength(dst_fta);
  ft_t *dst_ft  = fta_getft_array(dst_fta);
  ep_t src_beg, src_end, dst_beg, dst_end;
  int src_ix, dst_ix;

  for (src_ix = 0, dst_ix = 0; src_ix < src_nft && dst_ix < dst_nft;) {
    src_beg = src_ft[src_ix]->bt;
    src_end = src_ft[src_ix]->et;
    dst_beg = dst_ft[dst_ix]->bt;
    dst_end = dst_ft[dst_ix]->et;

    if (dst_beg <= src_beg && dst_end >= src_end) {
      dst_beg = dst_ft[dst_ix]->bt = max(dst_beg,src_beg);
      dst_end = dst_ft[dst_ix]->et = min(dst_end,src_end);
      dst_ft[dst_ix]->assigned = src_ft[src_ix]->assigned;
    }
    /* move forward */
    if (dst_beg == src_beg && dst_end == src_end) {
      dst_ix++;
      src_ix++;
    }
    else if ((src_beg >= dst_beg && src_end <= dst_end) ||
	     (src_beg <= dst_beg && src_end <= dst_end))
      src_ix++;
    else {
      /* 
	 As a fortunate side-effect of the logic,
	 we need to delete the restricted entry every time
	 we would otherwise increment the restricted index.
	 This is not obvious, but it works!
      */
      fta_deleteft(dst_fta,dst_ix);
      dst_nft = fta_getlength(dst_fta);
    }
  }
  while (dst_ix < dst_nft) {
    fta_deleteft(dst_fta,dst_ix);
    dst_nft = fta_getlength(dst_fta);
  }
}

int fta_add_stringtime(fta_t fta, char *time)
{
  ep_t start, stop;
  int assigned;

  lowercase(time);
  switch(*time) {
  case 'm':			/* Monday */
    if (strncmp(time,"mon",3)!= 0)
      return(-1);
    time += 3;
    if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
      return(-1);
    /* convert time within day to time within week */
    start += epoch_base[1];  stop += epoch_base[1];
    fta_add_time(fta,start,stop,assigned);
    break;
  case 't':			/* Tuesday/Thursday */
    switch (*(time+1)) {
    case 'h':			/* Thursday */
      if (strncmp(time,"thu",3)!= 0)
	return(-1);
      time += 3;
      if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
	return(-1);
      /* convert time within day to time within week */
      start += epoch_base[4];  stop += epoch_base[4];
      fta_add_time(fta,start,stop,assigned);
      break;
    case 'u':			/* Tuesday */
      if (strncmp(time,"tue",3)!= 0)
	return(-1);
      time += 3;
      if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
	return(-1);
      /* convert time within day to time within week */
      start += epoch_base[2];  stop += epoch_base[2];
      fta_add_time(fta,start,stop,assigned);
      break;
    default:			/* illegal */
      return(-1);
    }
    break;
  case 'w':			/* Wednesday */
    if (strncmp(time,"wed",3)!= 0)
      return(-1);
    time += 3;
    if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
      return(-1);
    /* convert time within day to time within week */
    start += epoch_base[3];  stop += epoch_base[3];
    fta_add_time(fta,start,stop,assigned);
    break;
  case 'f':			/* Friday */
    if (strncmp(time,"fri",3)!= 0)
      return(-1);
    time += 3;
    if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
      return(-1);
    /* convert time within day to time within week */
    start += epoch_base[5];  stop += epoch_base[5];
    fta_add_time(fta,start,stop,assigned);
    break;
  case 's':			/* Saturday/Sunday */
    switch (*(time+1)) {
    case 'a':			/* Saturday */
      if (strncmp(time,"sat",3)!= 0)
	return(-1);
      time += 3;
      if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
	return(-1);
      /* convert time within day to time within week */
      start += epoch_base[6];  stop += epoch_base[6];
      fta_add_time(fta,start,stop,assigned);
      break;
    case 'u':			/* Sunday */
      if (strncmp(time,"sun",3)!= 0)
	return(-1);
      time += 3;
      if (conv_string_interval24(time,&start,&stop,&assigned) == -1)
	return(-1);
      /* convert time within day to time within week */
      start += epoch_base[0];  stop += epoch_base[0];
      fta_add_time(fta,start,stop,assigned);
      break;
    default:			/* illegal */
      return(-1);
    }
    break;
  default:			/* illegal */
    return(-1);
  }
  return(0);
}

pi_t pi_create(char *name)
{
  pi_t pi = (pi_t) xcalloc(1,sizeof(pi_s));
  pi->name = (char *)xmalloc(strlen(name)+1);
  pi->fta = fta_create();	/* raw free time data */
  pi->rfta = fta_create();	/* restricted free time data */
  pi->sfta = fta_create();	/* solution free time data */
  strcpy(pi->name,name);

  return(pi);
}

pi_t pia_find_person(pia_t pia, char *name)
{
  register int i;
  pi_t pi;

  if (pia != (pia_t)0) {
    for (i = 0; i < pia->npi; i++) {
      pi = pia->pi[i];
      if (strcasecmp(pi->name,name)== 0)
	return(pi);
    }
  }
  return((pi_t)0);
}

pi_t pia_add_person(pia_t pia, char *name)
{
  pi_t pi = (pi_t)0;

  if (pia != (pia_t)0) {
    if (pia->npi >= pia->npi_limit) { /* need more array space */
      if (pia->pi == (pi_t *)0)
	pia->pi = (pi_t *)
	  xmalloc((pia->npi_limit + PIA_INCREMENT) * sizeof(pi_t));
      else
	pia->pi = (pi_t *) 
	  xrealloc((char *) pia->pi,
		   (pia->npi_limit + PIA_INCREMENT) * sizeof(pi_t));
      pia->npi_limit += PIA_INCREMENT;
    }
    
    pia->pi[pia->npi++] = pi = pi_create(name);
  }
    
  return(pi);
}

pia_t pia_create(void)
{
  pia_t pia = (pia_t)xcalloc(1,sizeof(pia_s));
  return(pia);
}

void pia_sort_on_rfta(pia_t pia)
{
  int done, i, pialen = pia->npi;
  pi_t *pi = pia->pi, tpi;
  
  done = 0;
  while (done == 0) {
    done = 1;
    for (i = 0; i < pialen - 1; i++) {
      if (pi[i]->rfta->nft > pi[i+1]->rfta->nft) {
	tpi = pi[i];
	pi[i] = pi[i+1];
	pi[i+1] = tpi;
	done = 0;
      }
    }
  }
}

void pia_sort(pia_t pia)
{
  int done, i, pialen = pia->npi;
  pi_t *pi = pia->pi, tpi;
  
  done = 0;
  while (done == 0) {
    done = 1;
    for (i = 0; i < pialen - 1; i++) {
      if (strcasecmp(pi[i]->name,pi[i+1]->name)>0) {
	tpi = pi[i];
	pi[i] = pi[i+1];
	pi[i+1] = tpi;
	done = 0;
      }
    }
  }
}

int conv_string_interval24(char *time, ep_t *start, ep_t *stop, int *assigned)
{
  ep_t  d1, d2;

  if (!isdigit(*time))
    return(-1);
  d1 = ((int) *time++) - '0';
  if (isdigit(*time))
    d1 = (d1 * 10) + ((int) *time++) - '0';

  if (d1 < 7)
    d1 += 12;

  if (*time++ != '-')
    return(-1);
  if (!isdigit(*time))
    return(-1);
  d2 = ((int) *time++) - '0';
  if (isdigit(*time))
    d2 = (d2 * 10) + ((int) *time++) - '0';
  
  if (d2 < 7)
    d2 += 12;

  if (*time == '*')
    *assigned = 1;
  else if (*time == '\0')
    *assigned = 0;
  else
    return(-1);

  if (d1 > d2)
    d2 += 12;
  if (d1 > 24 || d2 > 24)
    return(-1);
  *start = d1 * (60 * 60);	/* convert hours to seconds */
  *stop = d2 * (60 * 60);
  return(0);
}

char *conv_interval24_string(int twelve_hour, ep_t bt, ep_t et, char *buf)
{
  ep_t day, hour1, hour2;

  day = bt / SECS_PER_DAY;
  hour1 = (bt % SECS_PER_DAY) / (60 * 60);
  hour2 = (et % SECS_PER_DAY) / (60 * 60);
  /* convert to twelve hour format, if necessary */
  if (twelve_hour) {
    hour1 = (hour1 > 12 ? (hour1 - 12) : hour1);
    hour2 = (hour2 > 12 ? (hour2 - 12) : hour2);
  }
  sprintf(buf,"%s%02d-%02d",week_day[day],(int)hour1,(int)hour2);
  return(buf);
}
